@@ -186,12 +186,7 @@ class PathInput(QtWidgets.QWidget):
     def _clear_location(self) -> None:
         self.update_path(None)
 
-    @property
-    def status_tip(self) -> str:
-        return self.text.statusTip()
-
-    @status_tip.setter
-    def status_tip(self, msg: str) -> None:
+    def setStatusTip(self, msg: str) -> None:
         self.text.setStatusTip(msg)
 
     def on_path_change(self, fn: Callable[[Optional[Path]], None]) -> None:
