@@ -187,7 +187,7 @@ class EncryptTab(TabMixin):
             QtGui.QRegularExpressionValidator(QtCore.QRegularExpression(PACKAGE_SUFFIX))
         )
         output_location = PathInput(path=self.app_data.encrypt_output_location)
-        output_location.status_tip = "Destination folder for the encrypted package"
+        output_location.setStatusTip("Destination folder for the encrypted package")
 
         # Add actions
         output_suffix.editingFinished.connect(
