@@ -1,5 +1,5 @@
 from pathlib import Path
-from typing import Iterable, List
+from typing import Iterable, List, Optional, cast
 
 from sett import APP_NAME_SHORT
 from .component import (
@@ -13,9 +13,11 @@ from .pyside import QtWidgets, QtCore, QtGui, QAction, open_window
 
 
 class FileSelectionWidget(QtWidgets.QGroupBox):
-    """File selection widget"""
+    """File selection widget."""
 
-    def __init__(self, title, parent, name_filter=None):
+    def __init__(
+        self, title: str, parent: QtWidgets.QWidget, name_filter: Optional[str] = None
+    ):
         super().__init__(parent)
         self.path = str(Path.home())
         self.name_filter = name_filter
@@ -32,7 +34,7 @@ class FileSelectionWidget(QtWidgets.QGroupBox):
         self._create_layout()
         self.setAcceptDrops(True)
 
-    def _create_layout(self, allow_dirs=False):
+    def _create_layout(self, allow_dirs: bool = False) -> None:
         input_type_label = "files" + (" and/or directories" if allow_dirs else "")
         toolbar = ToolBar(f"Add {input_type_label}", self)
         for action in self._create_actions():
@@ -74,8 +76,15 @@ class FileSelectionWidget(QtWidgets.QGroupBox):
 
     def _create_remove_selected_action(self) -> QAction:
         def clear_selected() -> None:
-            indices = self.file_list_view.selectedIndexes()
-            for index in indices:
+
+            # NOTE: cast() is needed here as selectedIndexes() wrongly hints
+            #       that it returns List[int] when in fact it returns
+            #       Sequence[QtCore.QModelIndex].
+            selected_indexes = (
+                cast(QtCore.QModelIndex, x)
+                for x in self.file_list_view.selectedIndexes()
+            )
+            for index in selected_indexes:
                 self.file_list_model.removeRows(index.row(), 1)
             self.file_list_model.layoutChanged.emit()
 
@@ -102,13 +111,13 @@ class FileSelectionWidget(QtWidgets.QGroupBox):
     def _update_paths(self, paths: Iterable[str]) -> None:
         paths = set(filter(None, paths))
         if paths:
-            self.path = Path(next(iter(paths))).parent
+            self.path = str(Path(next(iter(paths))).parent)
         self.file_list_model.setStringList(
             sorted(set(self.file_list_model.stringList()) | paths)
         )
         self.file_list_model.layoutChanged.emit()
 
-    def _add_files(self):
+    def _add_files(self) -> None:
         dialog = QtWidgets.QFileDialog(self)
         dialog.setFileMode(QtWidgets.QFileDialog.ExistingFiles)
         dialog.setDirectory(str(self.path))
@@ -120,21 +129,21 @@ class FileSelectionWidget(QtWidgets.QGroupBox):
 
 
 class DirectoryAndFileSelectionWidget(FileSelectionWidget):
-    """File selection widget extension adding directory selection"""
+    """File selection widget extension adding directory selection."""
 
-    def __init__(self, title, parent):
+    def __init__(self, title: str, parent: QtWidgets.QWidget):
         super().__init__(title, parent)
 
-    def _create_layout(self, allow_dirs=True):
+    def _create_layout(self, allow_dirs: bool = True) -> None:
         super()._create_layout(allow_dirs=allow_dirs)
 
-    def _create_actions(self):
+    def _create_actions(self) -> List[QtWidgets.QAbstractButton]:
         actions = super()._create_actions()
         actions.insert(1, self._create_add_directory_button())
         return actions
 
-    def _create_add_directory_button(self):
-        def add_directory():
+    def _create_add_directory_button(self) -> QAction:
+        def add_directory() -> None:
             directory = QtWidgets.QFileDialog.getExistingDirectory(
                 self, "Select directory", str(self.path)
             )
@@ -150,7 +159,7 @@ class DirectoryAndFileSelectionWidget(FileSelectionWidget):
 class ArchiveOnlyFileSelectionWidget(FileSelectionWidget):
     """File selection widget extension for selecting ZIP and TAR archives only."""
 
-    def __init__(self, title, parent):
+    def __init__(self, title: str, parent: QtWidgets.QWidget):
         super().__init__(title, parent, name_filter="Archives (*.zip *.tar)")
 
     def dropEvent(self, event: QtGui.QDropEvent) -> None:
