@@ -20,7 +20,8 @@ def load_authority_key_threaded(app_data: AppData) -> None:
     show_warning = warning_callback("Certification Authority Key Loading Error")
     run_thread(
         load_authority_key,
-        f_kwargs=dict(report_config=app_data.config, sett_config=app_data.config),
+        f_kwargs=dict(sett_config=app_data.config),
         forward_errors=show_warning,
         signals=dict(warning=show_warning, result=update_authority_key_in_app_data),
+        report_config=app_data.config,
     )
