@@ -4,7 +4,7 @@ import sys
 import traceback
 import warnings
 from contextlib import contextmanager
-from typing import Sequence, Callable, Optional, Dict, cast
+from typing import Generator, Sequence, Callable, Optional, Dict, cast, Any
 
 from .pyside import QtCore, get_application_global_instance
 from ..core.secret import enforce_secret_by_signature
@@ -18,19 +18,30 @@ from ..utils.log import EmojiLevelFormatter, LOG_FORMAT, LOG_FORMAT_WITH_EMOJI
 
 
 class LoggerHandler(logging.Handler):
-    def __init__(self, level=logging.NOTSET, callback=None):
+    """Set color scheme for logging handler output."""
+
+    def __init__(
+        self,
+        level: int = logging.NOTSET,
+        callback: Optional[Callable[..., Any]] = None,
+    ):
         super().__init__(level)
         self.callback = callback
-        palette = get_application_global_instance().palette()
+        core_application = get_application_global_instance()
+        color = (
+            core_application.palette().color(core_application.palette().ColorRole.Text)
+            if core_application is not None
+            else "black"
+        )
         self.log_style = {
             logging.DEBUG: "gray",
-            logging.INFO: palette.color(palette.ColorRole.Text),
+            logging.INFO: color,
             logging.WARNING: "orange",
             logging.ERROR: "red",
             logging.CRITICAL: "red",
         }
 
-    def emit(self, record: logging.LogRecord):
+    def emit(self, record: logging.LogRecord) -> None:
         if self.callback:
             color = self.log_style.get(record.levelno)
             self.callback(
@@ -45,29 +56,28 @@ class WorkerSignals(QtCore.QObject):
     error = QtCore.Signal(tuple)
     result = QtCore.Signal(object)
     progress = QtCore.Signal(int)
-    logging = QtCore.Signal(str)
+    logging = cast(QtCore.SignalInstance, QtCore.Signal(str))
     warning = QtCore.Signal(str)
 
 
 class Worker(QtCore.QRunnable):
     """Worker can run an arbitrary function in a separate thread.
 
-    :param fn: function to run in a thread
-    :param args: arguments passed to the function
-    :param kwargs: keyword arguments passed to the function
-    :param logger: logs from those loggers are passed to the logging
-                   signal
+    :param fn: function to run in a thread.
+    :param args: arguments passed to the function.
+    :param kwargs: keyword arguments passed to the function.
+    :param logger: logs from those loggers are passed to the logging signal.
     """
 
     def __init__(
         self,
-        fn,
-        *args,
+        fn: Callable[..., Any],
+        *args: Any,
         capture_loggers: Sequence[logging.Logger] = (),
         ignore_exceptions: bool = False,
         forward_errors: Optional[Callable[[str], None]] = None,
         report_config: Config,
-        **kwargs,
+        **kwargs: Any,
     ):
         super().__init__()
 
@@ -121,7 +131,9 @@ class Worker(QtCore.QRunnable):
 
 
 @contextmanager
-def attach_loggers_to_signals(loggers: Sequence[logging.Logger], signals):
+def attach_loggers_to_signals(
+    loggers: Sequence[logging.Logger], signals: WorkerSignals
+) -> Generator[None, None, None]:
     logger_handler = LoggerHandler(callback=signals.logging.emit)
     logger_handler.setFormatter(
         logging.Formatter(LOG_FORMAT)
@@ -132,16 +144,17 @@ def attach_loggers_to_signals(loggers: Sequence[logging.Logger], signals):
     for logger in loggers:
         logger.addHandler(logger_handler)
     yield None
+
     for logger in loggers:
         logger.removeHandler(logger_handler)
 
 
 def run_thread(
-    f: Callable,
-    f_kwargs,
-    signals: Optional[Dict[str, Callable]] = None,
-    **worker_kwargs,
-):
+    f: Callable[..., Any],
+    f_kwargs: Any,
+    signals: Optional[Dict[str, Callable[..., Any]]] = None,
+    **worker_kwargs: Any,
+) -> None:
     pool = QtCore.QThreadPool.globalInstance()
     worker = Worker(f, **f_kwargs, **worker_kwargs)
     if signals is None:
