@@ -105,7 +105,9 @@ class TransferTab(TabMixin):
         bind(sftp_args, "jumphost", text_jumphost, OptionalTextControl)
 
         pkey_location = PathInput(directory=False, path=None)
-        pkey_location.status_tip = "Path to the private SSH key used for authentication"
+        pkey_location.setStatusTip(
+            "Path to the private SSH key used for authentication"
+        )
         bind(sftp_args, "pkey", pkey_location, OptionalPathControl)
 
         pkey_password = LineEdit()
