@@ -1,4 +1,4 @@
-from typing import Sequence, Optional
+from typing import Sequence, Optional, Callable, Any
 from abc import ABC, abstractmethod
 from ..utils.progress import ProgressInterface
 
@@ -8,8 +8,9 @@ class Protocol(ABC):
     def upload(
         self,
         files: Sequence[str],
+        two_factor_callback: Callable[[], str],
         progress: Optional[ProgressInterface] = None,
-    ):
+    ) -> Any:
         ...
 
     def required_password_args(self) -> Sequence[str]:  # pylint: disable=no-self-use
