@@ -2,8 +2,20 @@ import re
 import os
 import io
 from datetime import datetime
-from dataclasses import dataclass
-from typing import Optional, Callable, Tuple, Sequence
+from socket import socket
+from dataclasses import dataclass, field
+from typing import (
+    Optional,
+    Union,
+    Callable,
+    Tuple,
+    List,
+    Sequence,
+    Type,
+    Generator,
+    Any,
+    cast,
+)
 from pathlib import Path, PurePosixPath
 from contextlib import contextmanager
 
@@ -17,42 +29,59 @@ from paramiko import (
     ssh_exception,
     SSHClient,
     AutoAddPolicy,
-    SSHException,
     Agent,
+    Channel,
+    SFTPClient,
 )
+from paramiko.pkey import PKey
+from paramiko import SSHException as _SSHException
 
 from ..utils.progress import ProgressInterface
 from ..utils.log import create_logger
-from ..utils.config import get_config_file
+from ..utils.get_config_path import get_config_file
 from ..core.error import UserError
-from ..core.filesystem import to_human_readable_size
-from ..core.secret import Secret
+from ..core.filesystem import (
+    to_human_readable_size,
+    reverse_expanduser,
+    abspath_expanduser,
+)
+from ..core.secret import Secret, reveal
 from .defs import ENVELOPE_DIR_FMT
 from . import protocol
 
+# Re-exports
+SSHException = _SSHException
+
 
 logger = create_logger(__name__)
 
 
+def opt_reverse_expanduser(path: Optional[str]) -> Optional[str]:
+    return reverse_expanduser(path) if path else None
+
+
 @dataclass
 class Protocol(protocol.Protocol):
     host: str
     username: str
     destination_dir: str
-    two_factor_callback: Callable
     envelope_dir: Optional[str] = None
-    pkey: Optional[str] = None
-    pkey_password: Optional[Secret] = None
+    pkey: Optional[str] = field(
+        default=None,
+        metadata={
+            "serialize": opt_reverse_expanduser,
+            "deserialize": abspath_expanduser,
+        },
+    )
+    pkey_password: Optional[Secret[str]] = None
     pkey_password_encoding: str = (
         "utf_8"  # nosec (False Positive: pkey_password_encoding)
     )
     jumphost: Optional[str] = None
 
-    def __post_init__(self):
+    def __post_init__(self) -> None:
         # Turn empty strings into None and check that the ssh key file exists
         self.pkey = os.path.expanduser(self.pkey) if self.pkey else None
-        if self.pkey and not os.path.exists(self.pkey):
-            raise UserError(f"SSH key not found: {self.pkey}")
         self.pkey_password = self.pkey_password or None
 
     def required_password_args(self) -> Sequence[str]:
@@ -63,37 +92,47 @@ class Protocol(protocol.Protocol):
     def upload(
         self,
         files: Sequence[str],
+        two_factor_callback: Callable[[], str],
         progress: Optional[ProgressInterface] = None,
-    ):
+    ) -> None:
+        if self.pkey and not os.path.exists(self.pkey):
+            raise UserError(f"SSH key not found: {self.pkey}")
+
         if self.jumphost:
-            self.upload_paramiko(files, progress)
+            self.upload_paramiko(
+                files, progress=progress, two_factor_callback=two_factor_callback
+            )
         else:
             host, port = parse_host(self.host)
             try:
-                sett_rs.sftp_upload(  # pylint: disable=no-member
+                sett_rs.sftp_upload(
                     files,
                     f"{host}:{port}",
                     self.username,
                     self.destination_dir,
                     self.envelope_dir,
                     self.pkey,
-                    self.pkey_password,
-                    progress,
+                    reveal(self.pkey_password),
+                    progress.update if progress is not None else None,
+                    two_factor_callback=two_factor_callback,
                 )
             except RuntimeError as e:
-                logger.debug(
-                    "Upload failed with sett_rs: %s. Falling back to paramiko", str(e)
+                logger.warning(format(e))
+                self.upload_paramiko(
+                    files, progress=progress, two_factor_callback=two_factor_callback
                 )
-                self.upload_paramiko(files, progress)
 
     def upload_paramiko(
         self,
         files: Sequence[str],
+        two_factor_callback: Callable[[], str],
         progress: Optional[ProgressInterface] = None,
-    ):
+    ) -> None:
         if self.envelope_dir is None:
             envelope_dir = datetime.now().strftime(ENVELOPE_DIR_FMT)
-        progress_callback = progress and (lambda x, y: progress.update(x / y))
+        progress_callback = (
+            (lambda x, y: progress.update(x / y)) if progress is not None else None
+        )
         remote_dir = PurePosixPath(self.destination_dir) / envelope_dir
         try:
             with sftp_connection(
@@ -103,7 +142,7 @@ class Protocol(protocol.Protocol):
                 jumphost=self.jumphost,
                 pkey_password=self.pkey_password,
                 pkey_password_encoding=self.pkey_password_encoding,
-                two_factor_callback=self.two_factor_callback,
+                two_factor_callback=two_factor_callback,
             ) as sftp:
                 try:
                     sftp.mkdir(str(remote_dir))
@@ -159,27 +198,35 @@ class Protocol(protocol.Protocol):
 def sftp_connection(  # nosec
     host: str,
     username: str,
-    two_factor_callback: Callable,
+    two_factor_callback: Callable[[], str],
     pkey: Optional[str] = None,
-    pkey_password: Optional[Secret] = None,
+    pkey_password: Optional[Secret[str]] = None,
     pkey_password_encoding: str = "utf_8",
     jumphost: Optional[str] = None,
-):
-    key = pkey and private_key_from_file(
-        str(Path(pkey).expanduser()), pkey_password, encoding=pkey_password_encoding
+) -> Generator[SFTPClient, None, None]:
+    key = (
+        private_key_from_file(
+            str(Path(pkey).expanduser()), pkey_password, encoding=pkey_password_encoding
+        )
+        if pkey
+        else None
     )
     if jumphost is not None:
         pw = two_factor_callback()
-        sock = proxy_socket(host, jumphost, username, pkey=key, password=pw)
+        sock: Union[str, Channel] = proxy_socket(
+            host, jumphost, username, pkey=key, password=pw
+        )
     else:
         sock = host
-    trans = Transport(sock)
+    trans = Transport(cast(Union[str, socket], sock))
     trans.connect()
     try:
         auth(trans, username, key, two_factor_callback)
         sftp_client = None
         try:
             sftp_client = trans.open_sftp_client()
+            if sftp_client is None:
+                raise UserError(f"Could not open a sftp connection to {host}")
             yield sftp_client
         finally:
             if sftp_client:
@@ -188,9 +235,14 @@ def sftp_connection(  # nosec
         trans.close()
 
 
-def auth(trans, username, key, two_factor_callback: Callable):
-    allowed_types = set()
-    if key:
+def auth(
+    trans: Transport,
+    username: str,
+    key: Optional[PKey],
+    two_factor_callback: Callable[[], str],
+) -> None:
+    allowed_types = []
+    if key is not None:
         allowed_types = trans.auth_publickey(username, key)
     else:
         try:
@@ -204,7 +256,7 @@ def auth(trans, username, key, two_factor_callback: Callable):
         trans.auth_password(username, f2_code)
 
 
-def proxy_socket(host, jumphost, username, **kwargs):
+def proxy_socket(host: str, jumphost: str, username: str, **kwargs: Any) -> Channel:
     tunnel = SSHClient()
     tunnel.set_missing_host_key_policy(AutoAddPolicy())
     tunnel.connect(
@@ -213,7 +265,10 @@ def proxy_socket(host, jumphost, username, **kwargs):
         allow_agent=True,
         **kwargs,
     )
-    return tunnel.get_transport().open_channel(
+    transport = tunnel.get_transport()
+    if transport is None:
+        raise UserError(f"Could not open sftp channel to {host}")
+    return transport.open_channel(
         "direct-tcpip", parse_host(host), parse_host(jumphost)
     )
 
@@ -226,50 +281,51 @@ def parse_host(host: str) -> Tuple[str, int]:
         return host, 22
 
 
-def auth_handler(_title, _instructions, prompt_list):
+def auth_handler(_title: Any, _instructions: Any, prompt_list: Any) -> List[str]:
     if prompt_list:
         auth_url = re.findall(r"(https?://\S+)", prompt_list[0][0])
         if auth_url:
             logger.info("Authenticate at: %s", auth_url[0])
-    resp = ["" for _ in prompt_list]
-    return resp
+    return ["" for _ in prompt_list]
 
 
-def is_ascii(s):
+def is_ascii(s: str) -> bool:
     return all(ord(c) < 128 for c in s)
 
 
-def private_key_from_file(path, password, encoding="utf_8"):
-    errors = []
-    pass_bytes = None if password is None else password.encode(encoding)
+def private_key_from_file(
+    path: str, password: Optional[Secret[str]], encoding: str = "utf_8"
+) -> PKey:
+    errors = set()
+    pass_bytes = None if password is None else password.reveal().encode(encoding)
+    pkey_class: Type[PKey]
     for pkey_class in (RSAKey, DSSKey, ECDSAKey, Ed25519Key):
         try:
-            return pkey_class.from_private_key_file(path, pass_bytes)
+            return pkey_class.from_private_key_file(path, pass_bytes)  # type: ignore
         except (SSHException, ValueError) as e:
-            errors.append(e)
-    if password is not None and not is_ascii(password):
-        errors = [
-            (
-                "Your ssh secret key's password seems to contain "
-                "some non-ascii characters.\n"
-                "Either change your password ("
-                "`ssh-keygen -f <path to your private key> -p`)"
-                " or make sure the config option "
-                "`ssh_password_encoding` is set to the same "
-                "encoding, your key has been created with."
-                "Your config file is here:\n"
-                + get_config_file()
-                + "\nThe encoding is usually `utf_8` on linux "
-                "/ mac and `cp437` on windows for keys generated "
-                "with ssh-keygen"
-            )
-        ] + errors
+            errors.add(format(e).lower())
+    if password is not None and not is_ascii(password.reveal()):
+        errors.add(
+            "Your ssh secret key's password seems to contain "
+            "some non-ascii characters. "
+            "Either change your password ("
+            "`ssh-keygen -f <path to your private key> -p`)"
+            " or make sure the config option "
+            "`ssh_password_encoding` is set to the same "
+            "encoding that your key has been created with. "
+            f"Your config file is here: {get_config_file()}. "
+            "The encoding is usually `utf_8` on linux "
+            "/ mac and `cp437` on windows for keys generated "
+            "with ssh-keygen."
+        )
     raise UserError(
-        "Could not load private key:\n" + "\n".join(format(e) for e in errors)
+        "Could not load private key. "
+        f"Please make sure that the password for {path} is not empty and correct. "
+        f"Original errors: {'; '.join(sorted(errors, key=len))}."
     )
 
 
-def auth_from_agent(transport, username):
+def auth_from_agent(transport: Transport, username: str) -> List[str]:
     agent = Agent()
     try:
         for key in agent.get_keys():
