@@ -1,43 +1,29 @@
+import enum
+import json
 import os
 import platform
-import json
-import dataclasses
 import warnings
-import enum
 from dataclasses import dataclass, field as _field, fields
 from pathlib import Path
-from typing import Tuple, Dict, Any, Optional, Callable
+from typing import Dict, Optional, Callable, Union, cast, Any
 
-import gpg_lite as gpg
-from libbiomedit.lib.deserialize import deserialize, serialize
+from libbiomedit.lib import deserialize as serialization
 
-from .validation import REGEX_URL, PACKAGE_SUFFIX
 from .log import get_default_log_dir, create_logger
-from ..core.error import UserError
-from ..core.portal_api import PortalApi
+from .validation import REGEX_URL, PACKAGE_SUFFIX
+from .get_config_path import get_config_file, CONFIG_FILE_NAME
 from .. import APP_NAME_SHORT
+from ..core import gpg
+from ..core.error import UserError
+from ..core.filesystem import reverse_expanduser, abspath_expanduser
+from ..core.portal_api import PortalApi
+from ..protocols import Protocol, protocol_name, parse_protocol
 
-
-CONFIG_FILE_NAME = "config.json"
-CONFIG_FILE_ENVIRON_VAR = "SETT_CONFIG_FILE"
-
-conf_sub_dir_by_os: Dict[str, Tuple[str, ...]] = {
-    "Linux": (".config",),
-    "Darwin": (".config",),
-    "Windows": ("AppData", "Roaming"),
-}
+PKEY_PARAM_NAME = "pkey"
 
 logger = create_logger(__name__)
 
 
-@dataclass(frozen=True)
-class Connection:
-    """dataclass holding config of a connection (sftp / liquid files)"""
-
-    protocol: str
-    parameters: Dict[str, Any]
-
-
 class FileType(enum.Enum):
     file = enum.auto()
     directory = enum.auto()
@@ -53,33 +39,51 @@ class FieldMetadata:
     regex: Optional[str] = None
 
 
-def reverse_expanduser(path: str) -> str:
-    """Performs the reverse operation of os.path.expanduser(). It checks
-    whether a path starts with a user's home directory, and if so, changes
-    it to the "~" shortcut. It also removes any trailing path separator.
-
-    Example:
-      * On linux  : /home/user/foo/bar/ -> ~/foo/bar
-      * On windows: /home/user/foo/bar/ -> ~\\foo\\bar
-    """
-    path_as_posix = Path(path).as_posix()
-    home_as_posix = Path.home().as_posix()
-    if path_as_posix.startswith(home_as_posix):
-        return str(Path("~" + path_as_posix[len(home_as_posix) :]))
-
-    # Note: conversion to Path and back is to remove trailing path separators.
-    return str(Path(path))
-
-
 def reverse_expanduser_if_path_exists(path: Optional[str]) -> Optional[str]:
     return reverse_expanduser(path) if path else path
 
 
+SerializedConnection = Dict[str, Union[str, Dict[str, Any]]]
+
+
+def serialize_connection(connection: Protocol) -> SerializedConnection:
+    return {
+        "protocol": protocol_name[type(connection)],
+        "parameters": serialization.serialize(type(connection))(connection),
+    }
+
+
+def serialize_connections(
+    connections: Dict[str, Protocol]
+) -> Dict[str, SerializedConnection]:
+    return {
+        name: serialize_connection(connection)
+        for name, connection in connections.items()
+    }
+
+
+def deserialize_connections(
+    connections: Dict[str, SerializedConnection]
+) -> Dict[str, Protocol]:
+    def get_protocol(connection: SerializedConnection) -> str:
+        protocol = connection["protocol"]
+        if not isinstance(protocol, str):
+            raise ValueError("Invalid value for connection in config")
+        return protocol
+
+    return {
+        name: serialization.deserialize(parse_protocol(get_protocol(connection)))(
+            connection["parameters"]
+        )
+        for name, connection in connections.items()
+    }
+
+
 def field_ext(
-    serializer: Optional[Callable] = None,
-    deserializer: Optional[Callable] = None,
-    **kwargs,
-):
+    serialize: Optional[Callable[[Any], Any]] = None,
+    deserialize: Optional[Callable[[Any], Any]] = None,
+    **kwargs: Any,
+) -> Any:
     field_names = {f.name for f in fields(FieldMetadata)}
     field_kwargs = {key: val for key, val in kwargs.items() if key not in field_names}
     metadata = {
@@ -91,10 +95,10 @@ def field_ext(
         # The dataclasses doc specifies, that metadata should be a mapping.
         # Therefore we wrap our dataclass in a dict:
         field_kwargs["metadata"] = {"metadata": FieldMetadata(**metadata)}
-        if serializer:
-            field_kwargs["metadata"]["serialize"] = serializer
-        if deserializer:
-            field_kwargs["metadata"]["deserialize"] = deserializer
+        if serialize:
+            field_kwargs["metadata"]["serialize"] = serialize
+        if deserialize:
+            field_kwargs["metadata"]["deserialize"] = deserialize
     return _field(**field_kwargs)
 
 
@@ -123,8 +127,8 @@ class Config:
         # no validation, as specifying http:// is optional
     )
     gpg_home_dir: str = field_ext(
-        serializer=reverse_expanduser_if_path_exists,
-        deserializer=os.path.expanduser,
+        serialize=reverse_expanduser_if_path_exists,
+        deserialize=abspath_expanduser,
         default=gpg.get_default_gnupg_home_dir(),
         label="GPG home directory",
         description="Path of the directory where GnuPG stores its keyrings and "
@@ -166,8 +170,8 @@ class Config:
         "DTR verification and automatic PGP key downloading/updating is disabled.",
     )
     log_dir: str = field_ext(
-        serializer=reverse_expanduser_if_path_exists,
-        deserializer=os.path.expanduser,
+        serialize=reverse_expanduser_if_path_exists,
+        deserialize=abspath_expanduser,
         default=get_default_log_dir(),
         label="Log directory",
         description="Path to log files directory.",
@@ -184,10 +188,16 @@ class Config:
         description="Maximum number of logfiles to keep as backup. Set to 0 to disable logging.",
         minimum=0,
     )
-    connections: Dict[str, Connection] = _field(default_factory=dict)
+    connections: Dict[str, Protocol] = field_ext(
+        serialize=serialize_connections,
+        deserialize=deserialize_connections,
+        label="Connections",
+        description="List of used-defined connections for data transfer.",
+        default_factory=dict,
+    )
     output_dir: Optional[str] = field_ext(
-        serializer=reverse_expanduser_if_path_exists,
-        deserializer=os.path.expanduser,
+        serialize=reverse_expanduser_if_path_exists,
+        deserialize=abspath_expanduser,
         default=None,
         label="Output directory",
         description="Default output directory, relevant for encryption/decryption.",
@@ -243,18 +253,18 @@ class Config:
         f"{APP_NAME_SHORT} naming convention",
     )
 
-    def __post_init__(self):
+    def __post_init__(self) -> None:
         for url in ("dcc_portal_url", "repo_url"):
             setattr(
                 self, url, getattr(self, url).rstrip("/")  # pylint: disable=no-member
             )
 
     @property
-    def portal_api(self):
+    def portal_api(self) -> PortalApi:
         return PortalApi(self.dcc_portal_url)
 
     @property
-    def gpg_store(self):
+    def gpg_store(self) -> gpg.GPGStore:
         return open_gpg_dir(self.gpg_home_dir)
 
     @property
@@ -277,57 +287,36 @@ class Config:
 
 
 class ConnectionStore:
-    """Connection configuration storage manager"""
+    """Sftp/LiquidFiles connection configuration storage manager."""
 
     config_field_name = "connections"
 
     def __init__(self, config_path: Optional[str] = None):
         self.path = get_config_file() if config_path is None else config_path
 
-    def _read(self) -> dict:
+    def _read(self) -> Dict[str, Any]:
         """Load data from config file"""
-        try:
-            with open(self.path, encoding="utf-8") as f:
-                return json.load(f)
-        except FileNotFoundError:
-            return {}
+        return load_config_dict(self.path)
 
-    def _write(self, data: dict):
+    def _write(self, data: Dict[str, Any]) -> None:
         """Write data to config file"""
         save_config(data, self.path)
 
-    def save(self, name: str, connection: Connection, exclude_pattern: str = "pass"):
+    def save(self, name: str, connection: Protocol) -> None:
         """Save a new connection to the config file.
 
-        The "exclude_pattern" argument is used to remove all fields whose name
-        contain the specified string. The default is "pass", meaning that
-        fields named e.g. "password" or "passphrase" will be removed from the
-        Connection object before it gets written to a file.
-
         :param name: name of the new connection.
         :param connection: new connection object to write to the config file.
-        :param exclude_pattern: exclude all connection fields containing the
-            specified pattern.
         """
         # Load entire sett configuration file data as a dict. It might or might
         # not already contain a "connections" field containing data for one or
         # more connections. If not, an empty "connections" field is added.
         data = self._read()
         connections = data.setdefault(self.config_field_name, {})
-        connection_dict = dataclasses.asdict(connection)
-
-        # Note: in principle, the "pkey_password" parameter of a Connection
-        # object should not be present anymore at this point, but we
-        # nevertheless check to be sure.
-        for parameter_to_delete in {
-            k for k in connection_dict["parameters"] if exclude_pattern in k
-        }:
-            del connection_dict["parameters"][parameter_to_delete]
-
-        connections[name] = connection_dict
+        connections[name] = serialize_connection(connection)
         self._write(data)
 
-    def delete(self, name: str):
+    def delete(self, name: str) -> None:
         """Delete a new connection from the config file"""
         data = self._read()
         try:
@@ -336,7 +325,7 @@ class ConnectionStore:
         except KeyError as e:
             raise UserError(f"Connection '{name}' does not exist.") from e
 
-    def rename(self, old: str, new: str):
+    def rename(self, old: str, new: str) -> None:
         """Rename an existing connection from the config file"""
         data = self._read()
         try:
@@ -360,11 +349,15 @@ def load_config() -> Config:
     except UserError as e:
         warnings.warn(format(e))
 
-    return deserialize(Config)(cfg_dct)
+    return serialization.deserialize(Config)(cfg_dct)
 
 
-def save_config(config: dict, path: Optional[str] = None) -> str:
-    """Save config to a file.
+def save_config(config: Dict[str, Any], path: Optional[str] = None) -> str:
+    """Save the specified config values to the specified path (file name).
+    If no path is specified, the config file is saved in the default location.
+
+    :param config: application config values in the form of a dictionary.
+    :param path: path + name of the config file to write to disk.
     :return: path of the created config file.
     """
     config_file = Path(get_config_file() if path is None else path)
@@ -384,17 +377,18 @@ def create_config() -> str:
     return save_config(config_to_dict(default_config()))
 
 
-def config_to_dict(config: Config) -> dict:
+def config_to_dict(config: Config) -> Dict[str, Any]:
     """Converts a Config object into a dict."""
-    return serialize(Config)(config)
+    data: Dict[str, Any] = serialization.serialize(Config)(config)
+    return data
 
 
 def default_config() -> Config:
     """Creates a new Config object with default values."""
-    return deserialize(Config)(sys_config_dict())
+    return serialization.deserialize(Config)(sys_config_dict())
 
 
-def sys_config_dict() -> dict:
+def sys_config_dict() -> Dict[str, Any]:
     """On linux only: try to load global sys config. If the env variable
     SYSCONFIG is set search there, else in /etc/{APP_NAME_SHORT}.
     """
@@ -404,47 +398,22 @@ def sys_config_dict() -> dict:
     return {}
 
 
-def load_config_dict(path: str) -> dict:
+def load_config_dict(path: str) -> Dict[str, Any]:
     """Load raw config as a dict."""
     try:
         with open(path, encoding="utf-8") as f:
-            return json.load(f)
+            data: Dict[str, Any] = json.load(f)
+            if not isinstance(data, dict):
+                raise UserError(
+                    f"Failed to load configuration from '{path}'. Wrong format."
+                )
+            return data
     except FileNotFoundError:
         return {}
     except json.decoder.JSONDecodeError as e:
         raise UserError(f"Failed to load configuration from '{path}'. {e}") from e
 
 
-def get_config_file() -> str:
-    """Retrieve the platform-specific path of the config file. If the user has
-    the correct config file path environmental variable defined in their
-    current environment, this file gets used instead.
-
-    :return: path of the config.
-    :raises UserError:
-    """
-    # Case 1: a config file path environmental variable is defined.
-    if CONFIG_FILE_ENVIRON_VAR in os.environ:
-        config_file = os.environ[CONFIG_FILE_ENVIRON_VAR]
-        if os.path.isdir(config_file):
-            raise UserError(
-                f"Environmental variable {CONFIG_FILE_ENVIRON_VAR} "
-                f"must point to a file, not a directory "
-                f"[{config_file}]."
-            )
-        return config_file
-
-    # Case 2: use the default platform-specific config file.
-    return os.path.join(get_config_dir(), CONFIG_FILE_NAME)
-
-
-def get_config_dir() -> str:
-    """Return platform specific default config direcory."""
-    return os.path.join(
-        os.path.expanduser("~"), *conf_sub_dir_by_os[platform.system()], APP_NAME_SHORT
-    )
-
-
 def open_gpg_dir(gpg_dir: str) -> gpg.GPGStore:
     """Open the database inside a GnuPG directory and return it as a gpg
     object.
@@ -462,7 +431,7 @@ def open_gpg_dir(gpg_dir: str) -> gpg.GPGStore:
         raise UserError(f"unable to open GnuPG directory [{gpg_dir}].") from None
 
 
-def migrate_user_config_file():
+def migrate_user_config_file() -> None:
     """Temporary migration to auto-update user's config files after arguments
     are modified or deprecated in sett.
     """
@@ -482,7 +451,7 @@ def migrate_user_config_file():
         config_needs_update = True
 
     if "key_validation_authority_keyid" in config_dict:
-        fingerprint = Config().key_authority_fingerprint
+        fingerprint = cast(str, Config().key_authority_fingerprint)
         key_id = config_dict.pop("key_validation_authority_keyid")
         if key_id == fingerprint[-16:]:
             config_dict["key_authority_fingerprint"] = fingerprint
