@@ -1,10 +1,21 @@
 import io
 import os
 from contextlib import contextmanager
-from typing import List, Iterable, Tuple, Callable, Optional, Sequence, IO, cast
-from collections.abc import Collection
+from typing import (
+    List,
+    Iterable,
+    Generator,
+    Tuple,
+    Callable,
+    Optional,
+    Sequence,
+    IO,
+    cast,
+    Any,
+    Sized,
+    TypeVar,
+)
 from itertools import repeat
-from functools import wraps
 
 
 class ProgressInterface:
@@ -23,12 +34,15 @@ class ProgressInterface:
         """
 
 
+T = TypeVar("T")
+
+
 def progress_iter(
-    iterable: Iterable,
-    progress: ProgressInterface = None,
-    weights: List[float] = None,
-    iterable_length: int = None,
-) -> Iterable:
+    iterable: Iterable[T],
+    progress: Optional[ProgressInterface] = None,
+    weights: Optional[Sequence[float]] = None,
+    iterable_length: Optional[int] = None,
+) -> Iterable[T]:
     """Add progress tracking through a ProgressInterface to an iterable object.
     :param iterable: iterable object to which progress tracking is added.
     :param progress: progress interface objects that will take care of
@@ -58,11 +72,11 @@ def progress_iter(
 
 
 def progress_iter_with_sub_progress(
-    iterable: Iterable,
+    iterable: Iterable[T],
     progress: Optional[ProgressInterface] = None,
-    weights: List[float] = None,
-    iterable_length: int = None,
-) -> Iterable:
+    weights: Optional[Sequence[float]] = None,
+    iterable_length: Optional[int] = None,
+) -> Iterable[Tuple[T, Optional[ProgressInterface]]]:
     """Add progress tracking through a ProgressInterface to an iterable object.
     The difference with 'progress_iter' is that items of the returned iterable
     will now contain tuples consisting of the original item and a
@@ -87,28 +101,34 @@ def progress_iter_with_sub_progress(
     )
 
 
+R = TypeVar("R")
+
+
 def _progress_iter_common(
-    wrapper: Callable,
-    iterable: Iterable,
-    progress: ProgressInterface = None,
+    wrapper: Callable[
+        [ProgressInterface], Callable[[Tuple[T, Tuple[float, float]]], R]
+    ],
+    iterable: Iterable[T],
+    progress: ProgressInterface,
     weights: Optional[Sequence[float]] = None,
-    iterable_length: int = None,
-) -> Iterable:
+    iterable_length: Optional[int] = None,
+) -> Iterable[R]:
     """Iterate over all values in the input iterable and apply the specified
     wrapper function to them. The wrapper function must accept arguments that
     have the form (item, (start_position, stop)).
     """
-    # Get length of input iterable.
-    if isinstance(iterable, Collection):
-        iterable_length = iterable.__len__()
-    elif iterable_length is None:
-        raise ValueError(
-            "iterable has no length attribute and no value was"
-            "passed for parameter [iterable_length] either."
-        )
 
     # If no weight were provided, assume equal weights for all iterations.
     if weights is None:
+        if iterable_length is None:
+            try:
+                iterable_length = len(cast(Sized, iterable))
+            except:
+                raise ValueError(
+                    "Iterable has no length attribute and no value was"
+                    "passed for parameter [iterable_length] either."
+                ) from None
+
         weights = (1,) * iterable_length
 
     # Create sub-ranges of the 0-1 range proportionally to the input weights.
@@ -142,7 +162,9 @@ def weights_to_ranges(
         yield (start, stop)
 
 
-def _progress_wrapper(progress: ProgressInterface) -> Callable:
+def _progress_wrapper(
+    progress: ProgressInterface,
+) -> Callable[[Tuple[T, Tuple[float, float]]], T]:
     """Create a wrapper function that takes an 'item_with_range' input object
     and returns the bare input item (an iterable) and updates the progress
     interface with the 'start' position of the range associated to the item.
@@ -151,9 +173,7 @@ def _progress_wrapper(progress: ProgressInterface) -> Callable:
     (start, stop) giving the range of completion fractions for the iterable.
     """
 
-    def progress_wrapper(
-        item_with_range: Tuple[Iterable, Tuple[float, float]]
-    ) -> Iterable:
+    def progress_wrapper(item_with_range: Tuple[T, Tuple[float, float]]) -> T:
         """item_with_range: iterable and its completion range
         :return: The bare item
         """
@@ -164,15 +184,17 @@ def _progress_wrapper(progress: ProgressInterface) -> Callable:
     return progress_wrapper
 
 
-def _progress_wrapper_with_sub_progress(progress: ProgressInterface) -> Callable:
+def _progress_wrapper_with_sub_progress(
+    progress: ProgressInterface,
+) -> Callable[[Tuple[T, Tuple[float, float]]], Tuple[T, Optional[ProgressInterface]],]:
     """Return a function that takes an 'item_with_range' and return the
     iterable item with an associated ProgressInterface that does automatic
     rescaling of its .update() method in the range of 'item_with_range'.
     """
 
     def progress_wrapper(
-        item_with_range: Tuple[Iterable, Tuple[float, float]]
-    ) -> Tuple[Iterable, Optional[ProgressInterface]]:
+        item_with_range: Tuple[T, Tuple[float, float]]
+    ) -> Tuple[T, Optional[ProgressInterface]]:
         """item_with_range: iterable and its completion range
         :return: A tuple consisting of the bare item and a ProgressInterface instance
         """
@@ -184,7 +206,7 @@ def _progress_wrapper_with_sub_progress(progress: ProgressInterface) -> Callable
 
 
 def scale_progress(
-    progress: ProgressInterface, start: float, stop: float
+    progress: Optional[ProgressInterface], start: float, stop: float
 ) -> Optional[ProgressInterface]:
     """Create an instance of a ProgressInterface with a modified .update
     method that rescales the value of the 'completed_fraction' to be in the
@@ -193,13 +215,16 @@ def scale_progress(
     method of the new instance will rescale any 'completed_fraction' fraction
     value to the range 0.25 - 0.5. So calling .update(0.5) will in fact update
     the progress instance to 0.375 instead of 0.5, and update(1.0) will update
-     the progress instance to 0.5 and not to 1.0.
+    the progress instance to 0.5 and not to 1.0.
     """
     if start > stop:
         raise ValueError("start > stop")
     if progress is None:
         return None
 
+    # Narrow the type scope of the progress object for typing purposes.
+    scaled_progress: ProgressInterface = progress
+
     # Define a new class that will be returned by the function. The .update
     # method is modified so it automatically rescales its inputs to the
     # desired range of values.
@@ -209,20 +234,22 @@ def scale_progress(
             in the range [start-stop].
             """
             completed_fraction = min(completed_fraction, 1)
-            progress.update(start + completed_fraction * (stop - start))
+            scaled_progress.update(start + completed_fraction * (stop - start))
 
         def get_completed_fraction(self) -> float:
             """Rescale the completed fraction so that its value is in the
             range [start-stop].
             """
-            completed_fraction = progress.get_completed_fraction()
+            completed_fraction = scaled_progress.get_completed_fraction()
             return (completed_fraction - start) / (stop - start)
 
     return SubProgress()
 
 
 @contextmanager
-def subprogress(progress: Optional[ProgressInterface], step_completion_increase: float):
+def subprogress(
+    progress: Optional[ProgressInterface], step_completion_increase: float
+) -> Generator[Optional[ProgressInterface], None, None]:
     """Context manager for running a progress bar or a subset of a progress
     bar (i.e. a "scaled progress".
     Usage example:
@@ -255,69 +282,124 @@ def subprogress(progress: Optional[ProgressInterface], step_completion_increase:
     yield scale_progress(progress, start, stop)
 
 
-class FileObjectProgress:
+class FileObjectProgress(IO[bytes]):
     """Wraps a file object
 
     All read operations will be tracked and passed to the ProgressInterface
     instance
     """
 
-    def __init__(self, f_obj: IO[bytes], progress):
+    def __init__(self, f_obj: IO[bytes], progress: ProgressInterface):
         self._f = f_obj
         original_pos = f_obj.tell()
         f_obj.seek(0, io.SEEK_END)
         total_size = f_obj.tell() - original_pos
         f_obj.seek(original_pos)
+        self.progress = progress
 
         if total_size == 0:
 
-            def wrap_update(f):
+            def update_pos() -> None:
                 progress.update(1)
-                return f
 
         else:
 
-            def wrap_update(f):
-                @wraps(f)
-                def wrapped_f(*args, **kwargs):
-                    data = f(*args, **kwargs)
-                    progress.update((f_obj.tell() - original_pos) / total_size)
-                    return data
+            def update_pos() -> None:
+                progress.update((f_obj.tell() - original_pos) / total_size)
 
-                return wrapped_f
+        @contextmanager
+        def track_position() -> Generator[None, None, None]:
+            yield None
+            update_pos()
+
+        self.track_position = track_position
 
-        for f_name in ("__next__", "read", "readline", "readlines"):
-            f = getattr(f_obj, f_name)
-            if f is not None:
-                setattr(self, f_name, wrap_update(f))
         super().__init__()
 
-    def __iter__(self):
+    def __iter__(self) -> "FileObjectProgress":
         return self
 
-    def __next__(self):
-        # Note: self.__next__ != FileObjectProgress.__next__
-        return self.__next__()
+    def __next__(self) -> bytes:
+        with self.track_position():
+            return self._f.__next__()
 
-    def __enter__(self):
+    def __enter__(self) -> "FileObjectProgress":
         self._f.__enter__()
         return self
 
-    def __exit__(self, *args):
-        return self._f.__exit__(*args)
+    def __exit__(self, *args: Any) -> None:
+        self._f.__exit__(*args)
 
-    def __getattr__(self, attr):
-        return getattr(self._f, attr)
+    def close(self) -> None:
+        self._f.close()
 
     @property
-    def name(self):
-        return self.__getattr__("name")
+    def mode(self) -> str:
+        m: str = self._f.mode
+        return m
+
+    @property
+    def name(self) -> str:
+        return self._f.name
+
+    @property
+    def closed(self) -> bool:
+        return self._f.closed
+
+    def fileno(self) -> int:
+        return self._f.fileno()
+
+    def flush(self) -> None:
+        self._f.flush()
+
+    def isatty(self) -> bool:
+        return self._f.isatty()
+
+    def read(self, n: int = -1) -> bytes:
+        with self.track_position():
+            return self._f.read(n)
+
+    def readable(self) -> bool:
+        return self._f.readable()
+
+    def readline(self, limit: int = -1) -> bytes:
+        with self.track_position():
+            return self._f.readline(limit)
+
+    def readlines(self, hint: int = -1) -> List[bytes]:
+        with self.track_position():
+            return self._f.readlines(hint)
+
+    @staticmethod
+    def seek(offset: int, whence: int = 0) -> int:
+        raise RuntimeError("Trying to seek in a FileObjectProgress")
+
+    @staticmethod
+    def seekable() -> bool:
+        return False
+
+    def tell(self) -> int:
+        return self._f.tell()
+
+    def truncate(self, size: Optional[int] = None) -> int:
+        return self._f.truncate(size)
+
+    def writable(self) -> bool:
+        return self._f.writable()
+
+    @staticmethod
+    def write(s: bytes) -> int:
+        raise RuntimeError("Trying to write in a FileObjectProgress")
+
+    @staticmethod
+    def writelines(lines: Iterable[bytes]) -> None:
+        raise RuntimeError("Trying to write in a FileObjectProgress")
 
 
 def with_progress(f: IO[bytes], progress: Optional[ProgressInterface]) -> IO[bytes]:
     if progress is None:
         return f
-    return cast(IO[bytes], FileObjectProgress(f, progress))
+    return FileObjectProgress(f, progress)
 
 
 class Opener:
@@ -333,19 +415,22 @@ class Opener:
     the first loop
     """
 
-    def __init__(self, f: str, mode: str, wrapper=lambda x: x):
+    def __init__(
+        self, f: str, mode: str, wrapper: Callable[[IO[bytes]], IO[bytes]] = lambda x: x
+    ):
         self.name = f
         self.mode = mode
         self.encoding = None if "b" in self.mode else "utf-8"
-        self._f = None
+        self._f: Optional[IO[bytes]] = None
         self.wrapper = wrapper
 
-    def __enter__(self):
+    def __enter__(self) -> IO[bytes]:
         self._f = open(self.name, self.mode, encoding=self.encoding)
         return self.wrapper(self._f.__enter__())
 
-    def __exit__(self, *args):
-        return self._f.__exit__(*args)
+    def __exit__(self, *args: Any) -> None:
+        if self._f is not None:
+            self._f.__exit__(*args)
 
 
 def progress_file_iter(
@@ -375,8 +460,10 @@ def progress_file_iter(
         )
 
 
-def with_progress_decorator(progress):
-    def _wrapper(f):
+def with_progress_decorator(
+    progress: Optional[ProgressInterface],
+) -> Callable[[IO[bytes]], IO[bytes]]:
+    def _wrapper(f: IO[bytes]) -> IO[bytes]:
         return with_progress(f=f, progress=progress)
 
     return _wrapper
