@@ -3,6 +3,7 @@ import os
 from functools import partial
 from typing import List, Optional
 
+from ..core import gpg
 from ..core.archive import (
     check_package,
     extract,
@@ -19,7 +20,6 @@ from ..core.crypt import (
     retrieve_refresh_and_validate_keys,
     fingerprint2keyid,
     verify_metadata_signature,
-    gpg,
 )
 from ..core.filesystem import (
     DeleteDirOnError,
@@ -44,14 +44,15 @@ logger = create_logger(__name__)
 def decrypt(
     files: List[str],
     *,
-    passphrase: Optional[Secret] = None,
+    passphrase: Optional[Secret[str]] = None,
     output_dir: str,
     config: Config,
     decrypt_only: bool = False,
     dry_run: bool = False,
     progress: Optional[ProgressInterface] = None,
-):
-    """Decrypt and decompress an input .tar file."""
+) -> None:
+    """Main function of the decrypt workflow. Decrypts and decompresses the
+    input archive files."""
 
     logger.info(
         "File(s) to decrypt: [%s], output_dir: %s%s",
@@ -67,7 +68,7 @@ def decrypt(
     check_space(get_total_size(files), output_dir, force=False)
 
     if dry_run:
-        logger.info("Dry run completed successfully")
+        logger.info("Dry run completed successfully.")
         return
 
     for archive_file in files:
@@ -86,12 +87,14 @@ def decrypt(
 
 def decrypt_archive(
     archive_file: str,
-    passphrase: Optional[Secret],
+    passphrase: Optional[Secret[str]],
     output_dir: str,
     config: Config,
     decrypt_only: bool = False,
     progress: Optional[ProgressInterface] = None,
-):
+) -> None:
+    """Decrypts and decompresses the input archive_file."""
+
     logger.info("Processing file %s", archive_file)
     # To avoid overwriting files, each archive is unpacked in a directory
     # that has the same name as the archive file minus its extension.
@@ -201,7 +204,7 @@ def decrypt_archive(
             )
 
 
-def log_files(files: List[str]):
+def log_files(files: List[str]) -> None:
     max_files_to_list = 5
     d_n = len(files) - max_files_to_list
     logger.info(
