@@ -4,11 +4,10 @@ import json
 import os
 from functools import partial
 from pathlib import Path
-from typing import List, Optional, Callable, Tuple, Union, Iterable
+from typing import List, Optional, Callable, Tuple, Iterable, Any, TypeVar
 from zipfile import ZipFile, ZipInfo, ZIP_STORED
 
-from ..utils.config import Config
-from ..utils.log import create_logger, log_runtime_info, log_timing
+from ..core import gpg
 from ..core.archive import (
     write_tar,
     ArchiveInMemoryFile,
@@ -35,7 +34,6 @@ from ..core.crypt import (
     detach_sign_file,
     check_password,
     search_priv_key,
-    gpg,
 )
 from ..core.checksum import generate_checksums_file_content, compute_checksum_on_write
 from ..core.error import UserError
@@ -43,35 +41,34 @@ from ..core.metadata import MetaData, alnum_str, Purpose, HexStr1024, HexStr256
 from ..core.secret import Secret
 from ..core.portal_api import PortalApi
 from ..utils.progress import ProgressInterface, subprogress, progress_file_iter
+from ..utils.config import Config
+from ..utils.log import create_logger, log_runtime_info, log_timing
 
 DATE_FMT_FILENAME = "%Y%m%dT%H%M%S"
 logger = create_logger(__name__)
 output_name_str = alnum_str(min_length=1, max_length=50, allow_dots=True)
 
 
-def path_str(directory: bool = False, writable: bool = False) -> Callable:
+def path_str(directory: bool = False, writable: bool = False) -> Callable[[str], str]:
     """Generate a 'type definition' function that will check that a string is
     a valid path (file or directory).
+
     :param directory: if True, the path must be a directory.
     :param writable: if True, the user must have write access to the path.
     :returns: type check function.
     :raises ValueError:
     """
 
-    def _path_str(path_to_check):
+    def _path_str(path_to_check: str) -> str:
         path = Path(path_to_check)
         if not path.exists():
-            raise ValueError(
-                f"Invalid path: '{path_to_check}'. " f"Path does not exist."
-            )
+            raise ValueError(f"Invalid path: '{path_to_check}'. Path does not exist.")
         if directory and not path.is_dir():
             raise ValueError(
-                f"Invalid path: '{path_to_check}'. " f"Path is not a directory."
+                f"Invalid path: '{path_to_check}'. Path is not a directory."
             )
         if writable and not os.access(path.as_posix(), os.W_OK):
-            raise ValueError(
-                f"Invalid path: '{path_to_check}'. " f"Path is not writable."
-            )
+            raise ValueError(f"Invalid path: '{path_to_check}'. Path is not writable.")
         return path_to_check
 
     return _path_str
@@ -80,26 +77,24 @@ def path_str(directory: bool = False, writable: bool = False) -> Callable:
 output_path_str = path_str(directory=True, writable=True)
 
 
-def check_integer_in_range(min_value: int, max_value: int) -> Callable:
+def check_integer_in_range(min_value: int, max_value: int) -> Callable[[Any], None]:
     """Generate a function that will check that its input value is an integer,
     and that this integer is in the range [min_value:max_value].
     """
 
-    def _integer_in_range(value_to_check: Union[int, str]) -> None:
+    def _integer_in_range(value_to_check: Any) -> None:
         try:
             value_to_check = int(value_to_check)
         except ValueError as e:
-            raise ValueError("value must be an integer") from e
+            raise ValueError("Value must be an integer") from e
 
         if value_to_check < min_value or value_to_check > max_value:
-            raise ValueError(
-                f"value must be in the range " f"[{min_value}-{max_value}]"
-            )
+            raise ValueError(f"Value must be in the range: [{min_value}-{max_value}]")
 
     return _integer_in_range
 
 
-def check_paths_on_posix(paths: Iterable[str]):
+def check_paths_on_posix(paths: Iterable[str]) -> None:
     """Check paths for windows-style separators.
 
     If the machine is POSIX, verify that the specified paths do not contain
@@ -110,7 +105,7 @@ def check_paths_on_posix(paths: Iterable[str]):
         if os.path.sep == "/" and "\\" in path:
             raise UserError(
                 "On POSIX systems (this machine), backslashes are NOT allowed "
-                f"as path separators. Offending value is [{path}]"
+                f"as path separators. Offending value is [{path}]."
             )
 
 
@@ -126,7 +121,7 @@ def encrypt(  # pylint: disable=too-many-statements
     recipient: List[str],
     dtr_id: Optional[int] = None,
     sender: Optional[str] = None,
-    passphrase: Optional[Secret] = None,
+    passphrase: Optional[Secret[str]] = None,
     output: Optional[str] = None,
     output_suffix: Optional[str] = None,
     dry_run: bool = False,
@@ -134,11 +129,16 @@ def encrypt(  # pylint: disable=too-many-statements
     force: bool = False,
     compression_level: Optional[int] = None,
     purpose: Optional[Purpose] = None,
-    progress: ProgressInterface = None,
+    progress: Optional[ProgressInterface] = None,
 ) -> Optional[str]:
     """Compress and encrypt files and/or directories.
 
-    Returns the file name of the created package.
+    Main function of the encryption workflow. It compresses the input files
+    into a single archive file, which is then encrypted for the specified
+    recipient and signed by the specified sender.
+    Finally, the encrypted data is bundled with a metadata file in a single
+    .zip archive.
+    The function returns the file name of the created package.
     """
 
     # Retrive non-specified, optional argument values from config.
@@ -151,7 +151,7 @@ def encrypt(  # pylint: disable=too-many-statements
     check_arg_value(
         arg_value=compression_level,
         arg_name="compression level",
-        arg_type=check_compression_level,
+        arg_type_checker=check_compression_level,
     )
 
     if not files:
@@ -360,23 +360,28 @@ def encrypt(  # pylint: disable=too-many-statements
     return output_name
 
 
+T = TypeVar("T")
+
+
 def check_arg_value(
-    arg_value: Union[str, int], arg_name: str, arg_type: Callable
+    arg_value: T,
+    arg_name: str,
+    arg_type_checker: Callable[[T], Any],
 ) -> None:
     """Verify that the value of variable arg_value that is named arg_name is
     following type arg_type. arg_type is a function that does a check
     of the type of the variable.
     """
     try:
-        arg_type(arg_value)
+        arg_type_checker(arg_value)
     except ValueError as e:
-        raise UserError(f"Invalid value for argument '{arg_name}': {e}") from e
+        raise UserError(f"Invalid value for argument '{arg_name}': {e}.") from e
 
 
 def byte_encode_metadata(
     metadata: MetaData,
     gpg_store: gpg.GPGStore,
-    passphrase: Optional[Secret],
+    passphrase: Optional[Secret[str]],
     sender_pub_key: Optional[gpg.Key],
 ) -> Tuple[bytes, bytes]:
     metadata_bytes = json.dumps(MetaData.asdict(metadata), indent=4).encode()
@@ -422,11 +427,14 @@ def determine_project_code(
 
 
 def generate_output_archive_name(output_name: Optional[str], default: str) -> str:
-    """Define the path + name of the output tarball file of the encrypt
-    workflow. If output_name does not contain any path information, the
-    output directory is set to the current working directory.
+    """Generates the path + name of the output archive file of the encrypt
+    workflow.
+    If output_name does not contain any path information, the output directory
+    is set to the current working directory.
+
     :param output_name: name or path + name of output tarball.
-    :param default: default file name to use if output_name is None or output_name is a folder
+    :param default: default file name to use if output_name is None or
+        output_name is a folder
     :return: path and name of the output tarball file.
     :raises UserError:
     """
