@@ -11,8 +11,10 @@ logger = create_logger(__name__)
 
 
 @log_runtime_info(logger)
-def request_sigs(key_ids: Sequence[str], *, config=Config):
-    """Requests signatures"""
+def request_sigs(key_ids: Sequence[str], *, config: Config) -> None:
+    """Sends a request to sign the specified keys to the key validation
+    authority specified in the config file."""
+
     if config.offline:
         raise UserError("Requesting key signature is not possible in offline mode.")
     keys = frozenset(search_pub_key(k, config.gpg_store, sigs=False) for k in key_ids)
